@@ -40,8 +40,7 @@ using namespace pal;
 
 PointSet::PointSet()
 {
-  nbPoints = cHullSize = 0;
-  cHull = nullptr;
+  nbPoints = 0;
   type = -1;
 }
 
@@ -67,9 +66,8 @@ PointSet::PointSet( double aX, double aY )
   , xmax( aY )
   , ymin( aX )
   , ymax( aY )
-
 {
-  nbPoints = cHullSize = 1;
+  nbPoints = 1;
   x.resize( 1 );
   y.resize( 1 );
   x[0] = aX;
@@ -82,26 +80,11 @@ PointSet::PointSet( const PointSet &ps )
   , ymin( ps.ymin )
   , ymax( ps.ymax )
 {
-  int i;
-
   nbPoints = ps.nbPoints;
   x = ps.x;
   y = ps.y;
 
-  if ( ps.cHull )
-  {
-    cHullSize = ps.cHullSize;
-    cHull = new int[cHullSize];
-    for ( i = 0; i < cHullSize; i++ )
-    {
-      cHull[i] = ps.cHull[i];
-    }
-  }
-  else
-  {
-    cHull = nullptr;
-    cHullSize = 0;
-  }
+  convexHull = ps.convexHull;
 
   type = ps.type;
 
@@ -181,10 +164,32 @@ void PointSet::invalidateGeos()
   GEOSContextHandle_t geosctxt = QgsGeos::getGEOSHandler();
   if ( mOwnsGeom ) // delete old geometry if we own it
     GEOSGeom_destroy_r( geosctxt, mGeos );
-  GEOSPreparedGeom_destroy_r( geosctxt, mPreparedGeom );
   mOwnsGeom = false;
   mGeos = nullptr;
-  mPreparedGeom = nullptr;
+
+  if ( mPreparedGeom )
+  {
+    GEOSPreparedGeom_destroy_r( geosctxt, mPreparedGeom );
+    mPreparedGeom = nullptr;
+  }
+
+  if ( mGeosPreparedBoundary )
+  {
+    GEOSPreparedGeom_destroy_r( geosctxt, mGeosPreparedBoundary );
+    mGeosPreparedBoundary = nullptr;
+  }
+
+  if ( mMultipartPreparedGeos )
+  {
+    GEOSPreparedGeom_destroy_r( geosctxt, mMultipartPreparedGeos );
+    mMultipartPreparedGeos = nullptr;
+  }
+  if ( mMultipartGeos )
+  {
+    GEOSGeom_destroy_r( geosctxt, mMultipartGeos );
+    mMultipartGeos = nullptr;
+  }
+
   mLength = -1;
   mArea = -1;
 }
@@ -200,9 +205,24 @@ PointSet::~PointSet()
   }
   GEOSPreparedGeom_destroy_r( geosctxt, mPreparedGeom );
 
-  deleteCoords();
+  if ( mGeosPreparedBoundary )
+  {
+    GEOSPreparedGeom_destroy_r( geosctxt, mGeosPreparedBoundary );
+    mGeosPreparedBoundary = nullptr;
+  }
 
-  delete[] cHull;
+  if ( mMultipartPreparedGeos )
+  {
+    GEOSPreparedGeom_destroy_r( geosctxt, mMultipartPreparedGeos );
+    mMultipartPreparedGeos = nullptr;
+  }
+  if ( mMultipartGeos )
+  {
+    GEOSGeom_destroy_r( geosctxt, mMultipartGeos );
+    mMultipartGeos = nullptr;
+  }
+
+  deleteCoords();
 }
 
 void PointSet::deleteCoords()
@@ -215,7 +235,7 @@ std::unique_ptr<PointSet> PointSet::extractShape( int nbPtSh, int imin, int imax
 {
   int i, j;
 
-  std::unique_ptr<PointSet> newShape = qgis::make_unique< PointSet >();
+  std::unique_ptr<PointSet> newShape = std::make_unique< PointSet >();
   newShape->type = GEOS_POLYGON;
   newShape->nbPoints = nbPtSh;
   newShape->x.resize( newShape->nbPoints );
@@ -256,7 +276,7 @@ bool PointSet::containsPoint( double x, double y ) const
     GEOSCoordSeq_setY_r( geosctxt, seq, 0, y );
     geos::unique_ptr point( GEOSGeom_createPoint_r( geosctxt, seq ) );
 #endif
-    bool result = ( GEOSPreparedContainsProperly_r( geosctxt, preparedGeom(), point.get() ) == 1 );
+    const bool result = ( GEOSPreparedContainsProperly_r( geosctxt, preparedGeom(), point.get() ) == 1 );
 
     return result;
   }
@@ -274,88 +294,62 @@ bool PointSet::containsLabelCandidate( double x, double y, double width, double
   return GeomFunction::containsCandidate( preparedGeom(), x, y, width, height, alpha );
 }
 
-void PointSet::splitPolygons( QLinkedList<PointSet *> &inputShapes,
-                              QLinkedList<PointSet *> &outputShapes,
-                              double xrm, double yrm )
+QLinkedList<PointSet *> PointSet::splitPolygons( PointSet *inputShape, double labelWidth, double labelHeight )
 {
-  int i, j;
-
-  int nbp;
-  std::vector< double > x;
-  std::vector< double > y;
-
-  int *pts = nullptr;
-
-  int *cHull = nullptr;
-  int cHullSize;
-
-  double cp;
-  double bestcp = 0;
+  int j;
 
   double bestArea = 0;
-  double area;
 
-  double base;
-  double b, c;
-  double s;
-
-  int ihs;
-  int ihn;
-
-  int ips;
-  int ipn;
+  double b;
 
   int holeS = -1;  // hole start and end points
   int holeE = -1;
 
   int retainedPt = -1;
-  int pt = 0;
 
-  double labelArea = xrm * yrm;
+  const double labelArea = labelWidth * labelHeight;
 
-  PointSet *shape = nullptr;
+  QLinkedList<PointSet *> inputShapes;
+  inputShapes.push_back( inputShape );
+  QLinkedList<PointSet *> outputShapes;
 
   while ( !inputShapes.isEmpty() )
   {
-    shape = inputShapes.takeFirst();
+    PointSet *shape = inputShapes.takeFirst();
 
-    x = shape->x;
-    y = shape->y;
-    nbp = shape->nbPoints;
-    pts = new int[nbp];
-
-    for ( i = 0; i < nbp; i++ )
+    const std::vector< double > &x = shape->x;
+    const std::vector< double > &y = shape->y;
+    const int nbp = shape->nbPoints;
+    std::vector< int > pts( nbp );
+    for ( int i = 0; i < nbp; i++ )
     {
       pts[i] = i;
     }
 
-    // conpute convex hull
-    shape->cHullSize = GeomFunction::convexHullId( pts, x, y, nbp, shape->cHull );
-
-    cHull = shape->cHull;
-    cHullSize = shape->cHullSize;
+    // compute convex hull
+    shape->convexHull = GeomFunction::convexHullId( pts, x, y );
 
     bestArea = 0;
     retainedPt = -1;
 
     // lookup for a hole
-    for ( ihs = 0; ihs < cHullSize; ihs++ )
+    for ( std::size_t ihs = 0; ihs < shape->convexHull.size(); ihs++ )
     {
       // ihs->ihn => cHull'seg
-      ihn = ( ihs + 1 ) % cHullSize;
+      const std::size_t ihn = ( ihs + 1 ) % shape->convexHull.size();
 
-      ips = cHull[ihs];
-      ipn = ( ips + 1 ) % nbp;
-      if ( ipn != cHull[ihn] ) // next point on shape is not the next point on cHull => there is a hole here !
+      const int ips = shape->convexHull[ihs];
+      const int ipn = ( ips + 1 ) % nbp;
+      if ( ipn != shape->convexHull[ihn] ) // next point on shape is not the next point on cHull => there is a hole here !
       {
-        bestcp = 0;
-        pt = -1;
+        double bestcp = 0;
+        int pt = -1;
         // lookup for the deepest point in the hole
-        for ( i = ips; i != cHull[ihn]; i = ( i + 1 ) % nbp )
+        for ( int i = ips; i != shape->convexHull[ihn]; i = ( i + 1 ) % nbp )
         {
-          cp = std::fabs( GeomFunction::cross_product( x[cHull[ihs]], y[cHull[ihs]],
-                          x[cHull[ihn]], y[cHull[ihn]],
-                          x[i], y[i] ) );
+          const double cp = std::fabs( GeomFunction::cross_product( x[shape->convexHull[ihs]], y[shape->convexHull[ihs]],
+                                       x[shape->convexHull[ihn]], y[shape->convexHull[ihn]],
+                                       x[i], y[i] ) );
           if ( cp - bestcp > EPSILON )
           {
             bestcp = cp;
@@ -366,17 +360,17 @@ void PointSet::splitPolygons( QLinkedList<PointSet *> &inputShapes,
         if ( pt  != -1 )
         {
           // compute the ihs->ihn->pt triangle's area
-          base = GeomFunction::dist_euc2d( x[cHull[ihs]], y[cHull[ihs]],
-                                           x[cHull[ihn]], y[cHull[ihn]] );
+          const double base = GeomFunction::dist_euc2d( x[shape->convexHull[ihs]], y[shape->convexHull[ihs]],
+                              x[shape->convexHull[ihn]], y[shape->convexHull[ihn]] );
 
-          b = GeomFunction::dist_euc2d( x[cHull[ihs]], y[cHull[ihs]],
+          b = GeomFunction::dist_euc2d( x[shape->convexHull[ihs]], y[shape->convexHull[ihs]],
                                         x[pt], y[pt] );
 
-          c = GeomFunction::dist_euc2d( x[cHull[ihn]], y[cHull[ihn]],
-                                        x[pt], y[pt] );
+          const double c = GeomFunction::dist_euc2d( x[shape->convexHull[ihn]], y[shape->convexHull[ihn]],
+                           x[pt], y[pt] );
 
-          s = ( base + b + c ) / 2; // s = half perimeter
-          area = s * ( s - base ) * ( s - b ) * ( s - c );
+          const double s = ( base + b + c ) / 2; // s = half perimeter
+          double area = s * ( s - base ) * ( s - b ) * ( s - c );
           if ( area < 0 )
             area = -area;
 
@@ -403,12 +397,12 @@ void PointSet::splitPolygons( QLinkedList<PointSet *> &inputShapes,
     int ps = -1, pe = -1, fps = -1, fpe = -1;
     if ( retainedPt >= 0 && bestArea > labelArea ) // there is a hole so we'll cut the shape in two new shape (only if hole area is bigger than twice labelArea)
     {
-      c = std::numeric_limits<double>::max();
+      double c = std::numeric_limits<double>::max();
 
       // iterate on all shape points except points which are in the hole
       bool isValid;
       int k, l;
-      for ( i = ( cHull[holeE] + 1 ) % nbp; i != ( cHull[holeS] - 1 + nbp ) % nbp; i = j )
+      for ( int i = ( shape->convexHull[holeE] + 1 ) % nbp; i != ( shape->convexHull[holeS] - 1 + nbp ) % nbp; i = j )
       {
         j = ( i + 1 ) % nbp; // i->j is shape segment not in hole
 
@@ -467,7 +461,7 @@ void PointSet::splitPolygons( QLinkedList<PointSet *> &inputShapes,
           pointY = pty;
         }
 
-        for ( k = cHull[holeS]; k != cHull[holeE]; k = ( k + 1 ) % nbp )
+        for ( k = shape->convexHull[holeS]; k != shape->convexHull[holeE]; k = ( k + 1 ) % nbp )
         {
           l = ( k + 1 ) % nbp;
           if ( GeomFunction::isSegIntersects( x[retainedPt], y[retainedPt], pointX, pointY, x[k], y[k], x[l], y[l] ) )
@@ -489,7 +483,7 @@ void PointSet::splitPolygons( QLinkedList<PointSet *> &inputShapes,
       }  // for point which are not in hole
 
       // we will cut the shapeu in two new shapes, one from [retainedPoint] to [newPoint] and one form [newPoint] to [retainedPoint]
-      int imin = retainedPt;
+      const int imin = retainedPt;
       int imax = ( ( ( fps < retainedPt && fpe < retainedPt ) || ( fps > retainedPt && fpe > retainedPt ) ) ? std::min( fps, fpe ) : std::max( fps, fpe ) );
 
       int nbPtSh1, nbPtSh2; // how many points in new shapes ?
@@ -547,8 +541,84 @@ void PointSet::splitPolygons( QLinkedList<PointSet *> &inputShapes,
     {
       outputShapes.append( shape );
     }
-    delete[] pts;
   }
+  return outputShapes;
+}
+
+void PointSet::offsetCurveByDistance( double distance )
+{
+  if ( !mGeos )
+    createGeosGeom();
+
+  if ( !mGeos || type != GEOS_LINESTRING )
+    return;
+
+  GEOSContextHandle_t geosctxt = QgsGeos::getGEOSHandler();
+  GEOSGeometry *newGeos;
+  try
+  {
+    newGeos = GEOSOffsetCurve_r( geosctxt, mGeos, distance, 0, GEOSBUF_JOIN_MITRE, 2 );
+    if ( !newGeos )
+      return;
+
+    // happens sometime, if the offset curve self-intersects
+    if ( GEOSGeomTypeId_r( geosctxt, newGeos ) == GEOS_MULTILINESTRING )
+    {
+      // we keep the longest part
+      const int nParts = GEOSGetNumGeometries_r( geosctxt, newGeos );
+      double maximumLength = -1;
+      const GEOSGeometry *longestPart = nullptr;
+      for ( int i = 0; i < nParts; ++i )
+      {
+        const GEOSGeometry *part = GEOSGetGeometryN_r( geosctxt, newGeos, i );
+        double partLength = -1;
+        if ( GEOSLength_r( geosctxt, part, &partLength ) == 1 )
+        {
+          if ( partLength > maximumLength )
+          {
+            maximumLength = partLength;
+            longestPart = part;
+          }
+        }
+      }
+
+      if ( !longestPart )
+      {
+        // something is really wrong!
+        GEOSGeom_destroy_r( geosctxt, newGeos );
+        return;
+      }
+
+      geos::unique_ptr longestPartClone( GEOSGeom_clone_r( geosctxt, longestPart ) );
+      GEOSGeom_destroy_r( geosctxt, newGeos );
+      newGeos = longestPartClone.release();
+    }
+
+    const int newNbPoints = GEOSGeomGetNumPoints_r( geosctxt, newGeos );
+    const GEOSCoordSequence *coordSeq = GEOSGeom_getCoordSeq_r( geosctxt, newGeos );
+    std::vector< double > newX;
+    std::vector< double > newY;
+    newX.resize( newNbPoints );
+    newY.resize( newNbPoints );
+    for ( int i = 0; i < newNbPoints; i++ )
+    {
+      GEOSCoordSeq_getX_r( geosctxt, coordSeq, i, &newX[i] );
+      GEOSCoordSeq_getY_r( geosctxt, coordSeq, i, &newY[i] );
+    }
+    nbPoints = newNbPoints;
+    x = newX;
+    y = newY;
+  }
+  catch ( GEOSException &e )
+  {
+    qWarning( "GEOS exception: %s", e.what() );
+    QgsMessageLog::logMessage( QObject::tr( "Exception: %1" ).arg( e.what() ), QObject::tr( "GEOS" ) );
+    return;
+  }
+
+  invalidateGeos();
+  mGeos = newGeos;
+  mOwnsGeom = true;
 }
 
 void PointSet::extendLineByDistance( double startDistance, double endDistance, double smoothDistance )
@@ -569,13 +639,13 @@ void PointSet::extendLineByDistance( double startDistance, double endDistance, d
     double lastY = y0;
     for ( int i = 1; i < nbPoints; ++i )
     {
-      double thisX = x[i];
-      double thisY = y[i];
+      const double thisX = x[i];
+      const double thisY = y[i];
       const double thisSegmentLength = std::sqrt( ( thisX - lastX ) * ( thisX - lastX ) + ( thisY - lastY ) * ( thisY - lastY ) );
       distanceConsumed += thisSegmentLength;
       if ( distanceConsumed >= smoothDistance )
       {
-        double c = ( distanceConsumed - smoothDistance ) / thisSegmentLength;
+        const double c = ( distanceConsumed - smoothDistance ) / thisSegmentLength;
         x1 = lastX + c * ( thisX - lastX );
         y1 = lastY + c * ( thisY - lastY );
         break;
@@ -594,8 +664,8 @@ void PointSet::extendLineByDistance( double startDistance, double endDistance, d
 
   if ( endDistance > 0 )
   {
-    double xend0 = x[nbPoints - 1];
-    double yend0 = y[nbPoints - 1];
+    const double xend0 = x[nbPoints - 1];
+    const double yend0 = y[nbPoints - 1];
     double xend1 = x[nbPoints - 2];
     double yend1 = y[nbPoints - 2];
 
@@ -605,13 +675,13 @@ void PointSet::extendLineByDistance( double startDistance, double endDistance, d
     double lastY = y0;
     for ( int i = nbPoints - 2; i >= 0; --i )
     {
-      double thisX = x[i];
-      double thisY = y[i];
+      const double thisX = x[i];
+      const double thisY = y[i];
       const double thisSegmentLength = std::sqrt( ( thisX - lastX ) * ( thisX - lastX ) + ( thisY - lastY ) * ( thisY - lastY ) );
       distanceConsumed += thisSegmentLength;
       if ( distanceConsumed >= smoothDistance )
       {
-        double c = ( distanceConsumed - smoothDistance ) / thisSegmentLength;
+        const double c = ( distanceConsumed - smoothDistance ) / thisSegmentLength;
         xend1 = lastX + c * ( thisX - lastX );
         yend1 = lastY + c * ( thisY - lastY );
         break;
@@ -664,19 +734,19 @@ OrientedConvexHullBoundingBox PointSet::computeConvexHullOrientedBoundingBox( bo
   bbox[2] = std::numeric_limits<double>::lowest();
   bbox[3] = std::numeric_limits<double>::lowest();
 
-  for ( int i = 0; i < cHullSize; i++ )
+  for ( std::size_t i = 0; i < convexHull.size(); i++ )
   {
-    if ( x[cHull[i]] < bbox[0] )
-      bbox[0] = x[cHull[i]];
+    if ( x[convexHull[i]] < bbox[0] )
+      bbox[0] = x[convexHull[i]];
 
-    if ( x[cHull[i]] > bbox[2] )
-      bbox[2] = x[cHull[i]];
+    if ( x[convexHull[i]] > bbox[2] )
+      bbox[2] = x[convexHull[i]];
 
-    if ( y[cHull[i]] < bbox[1] )
-      bbox[1] = y[cHull[i]];
+    if ( y[convexHull[i]] < bbox[1] )
+      bbox[1] = y[convexHull[i]];
 
-    if ( y[cHull[i]] > bbox[3] )
-      bbox[3] = y[cHull[i]];
+    if ( y[convexHull[i]] > bbox[3] )
+      bbox[3] = y[convexHull[i]];
   }
 
   OrientedConvexHullBoundingBox finalBb;
@@ -717,15 +787,16 @@ OrientedConvexHullBoundingBox PointSet::computeConvexHullOrientedBoundingBox( bo
     bb[15] = bb[13] - d1; // hx, hy
 
     // adjust all points
-    for ( int i = 0; i < 16; i += 4 )
+    for ( int  i = 0; i < 16; i += 4 )
     {
 
       alpha_seg = ( ( i / 4 > 0 ? ( i / 4 ) - 1 : 3 ) ) * M_PI_2 + alpha;
 
       double best_cp = std::numeric_limits<double>::max();
-      for ( int j = 0; j < nbPoints; j++ )
+
+      for ( std::size_t j = 0; j < convexHull.size(); j++ )
       {
-        const double cp = GeomFunction::cross_product( bb[i + 2], bb[i + 3], bb[i], bb[i + 1], x[cHull[j]], y[cHull[j]] );
+        const double cp = GeomFunction::cross_product( bb[i + 2], bb[i + 3], bb[i], bb[i + 1], x[convexHull[j]], y[convexHull[j]] );
         if ( cp < best_cp )
         {
           best_cp = cp;
@@ -798,18 +869,37 @@ double PointSet::minDistanceToPoint( double px, double py, double *rx, double *r
     GEOSCoordSeq_setY_r( geosctxt, coord, 0, py );
     geos::unique_ptr geosPt( GEOSGeom_createPoint_r( geosctxt, coord ) );
 #endif
-    int type = GEOSGeomTypeId_r( geosctxt, mGeos );
+    const int type = GEOSGeomTypeId_r( geosctxt, mGeos );
     const GEOSGeometry *extRing = nullptr;
+#if GEOS_VERSION_MAJOR>3 || GEOS_VERSION_MINOR>=9
+    const GEOSPreparedGeometry *preparedExtRing = nullptr;
+#endif
+
     if ( type != GEOS_POLYGON )
     {
       extRing = mGeos;
+#if GEOS_VERSION_MAJOR>3 || GEOS_VERSION_MINOR>=9
+      preparedExtRing = preparedGeom();
+#endif
     }
     else
     {
       //for polygons, we want distance to exterior ring (not an interior point)
       extRing = GEOSGetExteriorRing_r( geosctxt, mGeos );
+#if GEOS_VERSION_MAJOR>3 || ( GEOS_VERSION_MAJOR == 3 && GEOS_VERSION_MINOR>=9 )
+      if ( ! mGeosPreparedBoundary )
+      {
+        mGeosPreparedBoundary = GEOSPrepare_r( geosctxt, extRing );
+      }
+      preparedExtRing = mGeosPreparedBoundary;
+#endif
     }
+
+#if GEOS_VERSION_MAJOR>3 || ( GEOS_VERSION_MAJOR == 3 && GEOS_VERSION_MINOR>=9 )
+    const geos::coord_sequence_unique_ptr nearestCoord( GEOSPreparedNearestPoints_r( geosctxt, preparedExtRing, geosPt.get() ) );
+#else
     geos::coord_sequence_unique_ptr nearestCoord( GEOSNearestPoints_r( geosctxt, extRing, geosPt.get() ) );
+#endif
     double nx;
     double ny;
 #if GEOS_VERSION_MAJOR>3 || GEOS_VERSION_MINOR>=8
@@ -898,12 +988,12 @@ void PointSet::getCentroid( double &px, double &py, bool forceInside ) const
 
 bool PointSet::boundingBoxIntersects( const PointSet *other ) const
 {
-  double x1 = ( xmin > other->xmin ? xmin : other->xmin );
-  double x2 = ( xmax < other->xmax ? xmax : other->xmax );
+  const double x1 = ( xmin > other->xmin ? xmin : other->xmin );
+  const double x2 = ( xmax < other->xmax ? xmax : other->xmax );
   if ( x1 > x2 )
     return false;
-  double y1 = ( ymin > other->ymin ? ymin : other->ymin );
-  double y2 = ( ymax < other->ymax ? ymax : other->ymax );
+  const double y1 = ( ymin > other->ymin ? ymin : other->ymin );
+  const double y2 = ( ymax < other->ymax ? ymax : other->ymax );
   return y1 <= y2;
 }
 
@@ -1011,3 +1101,51 @@ bool PointSet::isClosed() const
 {
   return qgsDoubleNear( x[0], x[nbPoints - 1] ) && qgsDoubleNear( y[0], y[nbPoints - 1] );
 }
+
+QString PointSet::toWkt() const
+{
+  if ( !mGeos )
+    createGeosGeom();
+
+  GEOSContextHandle_t geosctxt = QgsGeos::getGEOSHandler();
+
+  try
+  {
+    GEOSWKTWriter *writer = GEOSWKTWriter_create_r( geosctxt );
+
+    char *wkt = GEOSWKTWriter_write_r( geosctxt, writer, mGeos );
+    const QString res( wkt );
+
+    GEOSFree_r( geosctxt, wkt );
+
+    GEOSWKTWriter_destroy_r( geosctxt, writer );
+    writer = nullptr;
+
+    return res;
+  }
+  catch ( GEOSException &e )
+  {
+    qWarning( "GEOS exception: %s", e.what() );
+    QgsMessageLog::logMessage( QObject::tr( "Exception: %1" ).arg( e.what() ), QObject::tr( "GEOS" ) );
+    return QString();
+  }
+}
+
+std::tuple< std::vector< double >, double > PointSet::edgeDistances() const
+{
+  std::vector< double > distances( nbPoints );
+  double totalDistance = 0;
+  double oldX = -1.0, oldY = -1.0;
+  for ( int i = 0; i < nbPoints; i++ )
+  {
+    if ( i == 0 )
+      distances[i] = 0;
+    else
+      distances[i] = std::sqrt( std::pow( oldX - x[i], 2 ) + std::pow( oldY - y[i], 2 ) );
+
+    oldX = x[i];
+    oldY = y[i];
+    totalDistance += distances[i];
+  }
+  return std::make_tuple( std::move( distances ), totalDistance );
+}
