@@ -20,6 +20,7 @@
 
 #include "qgis_sip.h"
 #include "qgis_core.h"
+#include "qgis.h"
 
 #include <QDate>
 #include <QDateTime>
@@ -106,15 +107,15 @@ class QgsRange
      */
     bool contains( const QgsRange<T> &other ) const
     {
-      bool lowerOk = ( mIncludeLower && mLower <= other.mLower )
-                     || ( !mIncludeLower && mLower < other.mLower )
-                     || ( !mIncludeLower && !other.mIncludeLower && mLower <= other.mLower );
+      const bool lowerOk = ( mIncludeLower && mLower <= other.mLower )
+                           || ( !mIncludeLower && mLower < other.mLower )
+                           || ( !mIncludeLower && !other.mIncludeLower && mLower <= other.mLower );
       if ( !lowerOk )
         return false;
 
-      bool upperOk = ( mIncludeUpper && mUpper >= other.mUpper )
-                     || ( !mIncludeUpper && mUpper > other.mUpper )
-                     || ( !mIncludeUpper && !other.mIncludeUpper && mUpper >= other.mUpper );
+      const bool upperOk = ( mIncludeUpper && mUpper >= other.mUpper )
+                           || ( !mIncludeUpper && mUpper > other.mUpper )
+                           || ( !mIncludeUpper && !other.mIncludeUpper && mUpper >= other.mUpper );
       if ( !upperOk )
         return false;
 
@@ -126,13 +127,13 @@ class QgsRange
      */
     bool contains( T element ) const
     {
-      bool lowerOk = ( mIncludeLower && mLower <= element )
-                     || ( !mIncludeLower && mLower < element );
+      const bool lowerOk = ( mIncludeLower && mLower <= element )
+                           || ( !mIncludeLower && mLower < element );
       if ( !lowerOk )
         return false;
 
-      bool upperOk = ( mIncludeUpper && mUpper >= element )
-                     || ( !mIncludeUpper && mUpper > element );
+      const bool upperOk = ( mIncludeUpper && mUpper >= element )
+                           || ( !mIncludeUpper && mUpper > element );
       if ( !upperOk )
         return false;
 
@@ -167,7 +168,20 @@ class QgsRange
       return false;
     }
 
-  private:
+    bool operator==( const QgsRange<T> &other ) const
+    {
+      return mLower == other.mLower &&
+             mUpper == other.mUpper &&
+             mIncludeLower == other.includeLower() &&
+             mIncludeUpper == other.includeUpper();
+    }
+
+    bool operator!=( const QgsRange<T> &other ) const
+    {
+      return ( ! operator==( other ) );
+    }
+
+  protected:
 
     T mLower;
     T mUpper;
@@ -178,15 +192,88 @@ class QgsRange
 
 
 /**
- * QgsRange which stores a range of double values.
+ * \brief QgsRange which stores a range of double values.
+ * \ingroup core
  * \see QgsIntRange
  * \see QgsDateRange
  * \see QgsDateTimeRange
  * \since QGIS 3.0
  */
-typedef QgsRange< double > QgsDoubleRange;
+class CORE_EXPORT QgsDoubleRange : public QgsRange< double >
+{
+  public:
 
+#ifndef SIP_RUN
 
+    /**
+     * Constructor for QgsDoubleRange. The \a lower and \a upper bounds are specified,
+     * and optionally whether or not these bounds are included in the range.
+     *
+     * The default values for \a lower and \a upper construct an infinite range (see isInfinite()).
+     *
+     * \since QGIS 3.18
+     */
+    QgsDoubleRange( double lower = std::numeric_limits< double >::lowest(),
+                    double upper = std::numeric_limits< double >::max(),
+                    bool includeLower = true, bool includeUpper = true )
+      : QgsRange( lower, upper, includeLower, includeUpper )
+    {}
+#else
+
+    /**
+     * Constructor for QgsDoubleRange. The \a lower and \a upper bounds are specified,
+     * and optionally whether or not these bounds are included in the range.
+     */
+    QgsDoubleRange( double lower,
+                    double upper,
+                    bool includeLower = true, bool includeUpper = true )
+      : QgsRange( lower, upper, includeLower, includeUpper )
+    {}
+
+    /**
+     * Constructor for QgsDoubleRange containing an infinite range (see isInfinite()).
+     *
+     * \since QGIS 3.18
+     */
+    QgsDoubleRange()
+      : QgsRange( std::numeric_limits< double >::lowest(), std::numeric_limits< double >::max(), true, true )
+    {}
+#endif
+
+    /**
+     * Returns TRUE if the range consists of all possible values.
+     * \since QGIS 3.18
+     */
+    bool isInfinite() const
+    {
+      return lower() == std::numeric_limits< double >::lowest() && upper() == std::numeric_limits< double >::max();
+    }
+
+#ifdef SIP_RUN
+    SIP_PYOBJECT __repr__();
+    % MethodCode
+    QString str = QStringLiteral( "<QgsDoubleRange: %1%2, %3%4>" ).arg( sipCpp->includeLower() ? QStringLiteral( "[" ) : QStringLiteral( "(" ) )
+                  .arg( sipCpp->lower() )
+                  .arg( sipCpp->upper() )
+                  .arg( sipCpp->includeUpper() ? QStringLiteral( "]" ) : QStringLiteral( ")" ) );
+    sipRes = PyUnicode_FromString( str.toUtf8().constData() );
+    % End
+#endif
+
+    bool operator==( const QgsDoubleRange &other ) const
+    {
+      return qgsDoubleNear( mLower, other.mLower ) &&
+             qgsDoubleNear( mUpper, other.mUpper ) &&
+             mIncludeLower == other.includeLower() &&
+             mIncludeUpper == other.includeUpper();
+    }
+
+    bool operator!=( const QgsDoubleRange &other ) const
+    {
+      return ( ! operator==( other ) );
+    }
+
+};
 
 
 /**
@@ -197,7 +284,68 @@ typedef QgsRange< double > QgsDoubleRange;
  * \see QgsDateTimeRange
  * \since QGIS 3.0
  */
-typedef QgsRange< int > QgsIntRange;
+class CORE_EXPORT QgsIntRange : public QgsRange< int >
+{
+  public:
+
+#ifndef SIP_RUN
+
+    /**
+     * Constructor for QgsIntRange. The \a lower and \a upper bounds are specified,
+     * and optionally whether or not these bounds are included in the range.
+     *
+     * The default values for \a lower and \a upper construct an infinite range (see isInfinite()).
+     *
+     * \since QGIS 3.18
+     */
+    QgsIntRange( int lower = std::numeric_limits< int >::lowest(),
+                 int upper = std::numeric_limits< int >::max(),
+                 bool includeLower = true, bool includeUpper = true )
+      : QgsRange( lower, upper, includeLower, includeUpper )
+    {}
+#else
+
+    /**
+     * Constructor for QgsIntRange. The \a lower and \a upper bounds are specified,
+     * and optionally whether or not these bounds are included in the range.
+     */
+    QgsIntRange( int lower,
+                 int upper,
+                 bool includeLower = true, bool includeUpper = true )
+      : QgsRange( lower, upper, includeLower, includeUpper )
+    {}
+
+    /**
+     * Constructor for QgsIntRange containing an infinite range (see isInfinite()).
+     *
+     * \since QGIS 3.18
+     */
+    QgsIntRange()
+      : QgsRange( std::numeric_limits< int >::lowest(), std::numeric_limits< int >::max(), true, true )
+    {}
+#endif
+
+    /**
+     * Returns TRUE if the range consists of all possible values.
+     * \since QGIS 3.18
+     */
+    bool isInfinite() const
+    {
+      return lower() == std::numeric_limits< int >::lowest() && upper() == std::numeric_limits< int >::max();
+    }
+
+#ifdef SIP_RUN
+    SIP_PYOBJECT __repr__();
+    % MethodCode
+    QString str = QStringLiteral( "<QgsIntRange: %1%2, %3%4>" ).arg( sipCpp->includeLower() ? QStringLiteral( "[" ) : QStringLiteral( "(" ) )
+                  .arg( sipCpp->lower() )
+                  .arg( sipCpp->upper() )
+                  .arg( sipCpp->includeUpper() ? QStringLiteral( "]" ) : QStringLiteral( ")" ) );
+    sipRes = PyUnicode_FromString( str.toUtf8().constData() );
+    % End
+#endif
+
+};
 
 
 /**
@@ -316,9 +464,9 @@ class QgsTemporalRange
 
       if ( mLower.isValid() )
       {
-        bool lowerOk = ( mIncludeLower && mLower <= other.mLower )
-                       || ( !mIncludeLower && mLower < other.mLower )
-                       || ( !mIncludeLower && !other.mIncludeLower && mLower <= other.mLower );
+        const bool lowerOk = ( mIncludeLower && mLower <= other.mLower )
+                             || ( !mIncludeLower && mLower < other.mLower )
+                             || ( !mIncludeLower && !other.mIncludeLower && mLower <= other.mLower );
         if ( !lowerOk )
           return false;
       }
@@ -328,9 +476,9 @@ class QgsTemporalRange
 
       if ( mUpper.isValid() )
       {
-        bool upperOk = ( mIncludeUpper && mUpper >= other.mUpper )
-                       || ( !mIncludeUpper && mUpper > other.mUpper )
-                       || ( !mIncludeUpper && !other.mIncludeUpper && mUpper >= other.mUpper );
+        const bool upperOk = ( mIncludeUpper && mUpper >= other.mUpper )
+                             || ( !mIncludeUpper && mUpper > other.mUpper )
+                             || ( !mIncludeUpper && !other.mIncludeUpper && mUpper >= other.mUpper );
         if ( !upperOk )
           return false;
       }
@@ -348,16 +496,16 @@ class QgsTemporalRange
 
       if ( mLower.isValid() )
       {
-        bool lowerOk = ( mIncludeLower && mLower <= element )
-                       || ( !mIncludeLower && mLower < element );
+        const bool lowerOk = ( mIncludeLower && mLower <= element )
+                             || ( !mIncludeLower && mLower < element );
         if ( !lowerOk )
           return false;
       }
 
       if ( mUpper.isValid() )
       {
-        bool upperOk = ( mIncludeUpper && mUpper >= element )
-                       || ( !mIncludeUpper && mUpper > element );
+        const bool upperOk = ( mIncludeUpper && mUpper >= element )
+                             || ( !mIncludeUpper && mUpper > element );
         if ( !upperOk )
           return false;
       }
@@ -451,6 +599,51 @@ class QgsTemporalRange
       return changed;
     }
 
+#ifndef SIP_RUN
+
+    /**
+     * Merges a list of temporal ranges.
+     *
+     * Any overlapping ranges will be converted to a single range which covers the entire
+     * range of the input ranges.
+     *
+     * The returned value will be a list of non-contiguous ranges which completely encompass
+     * the input \a ranges, sorted in ascending order.
+     *
+     * \note Not available in Python bindings
+     *
+     * \since QGIS 3.20
+     */
+    static QList< QgsTemporalRange<T> > mergeRanges( const QList< QgsTemporalRange<T> > &ranges )
+    {
+      if ( ranges.empty() )
+        return {};
+
+      QList< QgsTemporalRange<T > > sortedRanges = ranges;
+      std::sort( sortedRanges.begin(), sortedRanges.end(), []( const QgsTemporalRange< T > &a, const QgsTemporalRange< T > &b ) -> bool { return a.begin() < b.begin(); } );
+      QList< QgsTemporalRange<T>> res;
+      res.reserve( sortedRanges.size() );
+
+      QgsTemporalRange<T> prevRange;
+      auto it = sortedRanges.constBegin();
+      prevRange = *it++;
+      for ( ; it != sortedRanges.constEnd(); ++it )
+      {
+        if ( prevRange.overlaps( *it ) )
+        {
+          prevRange.extend( *it );
+        }
+        else
+        {
+          res << prevRange;
+          prevRange = *it;
+        }
+      }
+      res << prevRange;
+      return res;
+    }
+#endif
+
     bool operator==( const QgsTemporalRange<T> &other ) const
     {
       return mLower == other.mLower &&
@@ -485,6 +678,8 @@ class QgsTemporalRange
  */
 typedef QgsTemporalRange< QDate > QgsDateRange SIP_DOC_TEMPLATE;
 
+Q_DECLARE_METATYPE( QgsDateRange )
+
 /**
  * QgsRange which stores a range of date times.
  *
@@ -497,4 +692,6 @@ typedef QgsTemporalRange< QDate > QgsDateRange SIP_DOC_TEMPLATE;
  */
 typedef QgsTemporalRange< QDateTime > QgsDateTimeRange SIP_DOC_TEMPLATE;
 
+Q_DECLARE_METATYPE( QgsDateTimeRange )
+
 #endif // QGSRANGE_H
