@@ -22,6 +22,7 @@
 #include "qgsmessagelog.h"
 #include "qgssettings.h"
 #include "qgsexception.h"
+#include "qgsgeometryengine.h"
 
 #include <QElapsedTimer>
 #include <QObject>
@@ -70,11 +71,11 @@ QgsPostgresFeatureIterator::QgsPostgresFeatureIterator( QgsPostgresFeatureSource
     return;
   }
 
+  bool limitAtProvider = ( mRequest.limit() >= 0 );
+
   mCursorName = mConn->uniqueCursorName();
   QString whereClause;
 
-  bool limitAtProvider = ( mRequest.limit() >= 0 );
-
   bool useFallbackWhereClause = false;
   QString fallbackWhereClause;
 
@@ -83,6 +84,38 @@ QgsPostgresFeatureIterator::QgsPostgresFeatureIterator( QgsPostgresFeatureSource
     whereClause = whereClauseRect();
   }
 
+  // prepare spatial filter geometries for optimal speed
+  switch ( mRequest.spatialFilterType() )
+  {
+    case Qgis::SpatialFilterType::NoFilter:
+    case Qgis::SpatialFilterType::BoundingBox:
+      break;
+
+    case Qgis::SpatialFilterType::DistanceWithin:
+      // we only need to test the distance locally if we are transforming features on QGIS side, otherwise
+      // we use ST_DWithin on the postgres backend instead
+      if ( !mRequest.referenceGeometry().isEmpty() )
+      {
+        if ( !mTransform.isShortCircuited() || mSource->mSpatialColType != SctGeometry )
+        {
+          mDistanceWithinGeom = mRequest.referenceGeometry();
+          mDistanceWithinEngine.reset( QgsGeometry::createGeometryEngine( mDistanceWithinGeom.constGet() ) );
+          mDistanceWithinEngine->prepareGeometry();
+          limitAtProvider = false;
+        }
+        else
+        {
+          // we can safely hand this off to the backend to evaluate, so that it will nicely handle it within the query planner!
+          whereClause = QgsPostgresUtils::andWhereClauses( whereClause, QStringLiteral( "ST_DWithin(%1,ST_GeomFromText('%2',%3),%4)" ).arg(
+                          QgsPostgresConn::quotedIdentifier( mSource->mGeometryColumn ),
+                          mRequest.referenceGeometry().asWkt(),
+                          mSource->mRequestedSrid.isEmpty() ? mSource->mDetectedSrid : mSource->mRequestedSrid )
+                        .arg( mRequest.distanceWithin() ) );
+        }
+      }
+      break;
+  }
+
   if ( !mSource->mSqlWhereClause.isEmpty() )
   {
     whereClause = QgsPostgresUtils::andWhereClauses( whereClause, '(' + mSource->mSqlWhereClause + ')' );
@@ -113,23 +146,16 @@ QgsPostgresFeatureIterator::QgsPostgresFeatureIterator( QgsPostgresFeatureSource
     }
     mFilterRequiresGeometry = request.filterExpression()->needsGeometry();
 
-    if ( QgsSettings().value( QStringLiteral( "qgis/compileExpressions" ), true ).toBool() )
-    {
-      //IMPORTANT - this MUST be the last clause added!
-      QgsPostgresExpressionCompiler compiler = QgsPostgresExpressionCompiler( source );
+    //IMPORTANT - this MUST be the last clause added!
+    QgsPostgresExpressionCompiler compiler = QgsPostgresExpressionCompiler( source, request.flags() & QgsFeatureRequest::IgnoreStaticNodesDuringExpressionCompilation );
 
-      if ( compiler.compile( request.filterExpression() ) == QgsSqlExpressionCompiler::Complete )
-      {
-        useFallbackWhereClause = true;
-        fallbackWhereClause = whereClause;
-        whereClause = QgsPostgresUtils::andWhereClauses( whereClause, compiler.result() );
-        mExpressionCompiled = true;
-        mCompileStatus = Compiled;
-      }
-      else
-      {
-        limitAtProvider = false;
-      }
+    if ( compiler.compile( request.filterExpression() ) == QgsSqlExpressionCompiler::Complete )
+    {
+      useFallbackWhereClause = true;
+      fallbackWhereClause = whereClause;
+      whereClause = QgsPostgresUtils::andWhereClauses( whereClause, compiler.result() );
+      mExpressionCompiled = true;
+      mCompileStatus = Compiled;
     }
     else
     {
@@ -253,79 +279,88 @@ bool QgsPostgresFeatureIterator::fetchFeature( QgsFeature &feature )
   if ( mClosed )
     return false;
 
-  if ( mFeatureQueue.empty() && !mLastFetch )
+  while ( true )
   {
+    if ( mFeatureQueue.empty() && !mLastFetch )
+    {
 #if 0 //disabled dynamic queue size
-    QElapsedTimer timer;
-    timer.start();
+      QElapsedTimer timer;
+      timer.start();
 #endif
 
-    QString fetch = QStringLiteral( "FETCH FORWARD %1 FROM %2" ).arg( mFeatureQueueSize ).arg( mCursorName );
-    QgsDebugMsgLevel( QStringLiteral( "fetching %1 features." ).arg( mFeatureQueueSize ), 4 );
+      QString fetch = QStringLiteral( "FETCH FORWARD %1 FROM %2" ).arg( mFeatureQueueSize ).arg( mCursorName );
+      QgsDebugMsgLevel( QStringLiteral( "fetching %1 features." ).arg( mFeatureQueueSize ), 4 );
 
-    lock();
-    if ( mConn->PQsendQuery( fetch ) == 0 ) // fetch features asynchronously
-    {
-      QgsMessageLog::logMessage( QObject::tr( "Fetching from cursor %1 failed\nDatabase error: %2" ).arg( mCursorName, mConn->PQerrorMessage() ), QObject::tr( "PostGIS" ) );
-    }
-
-    QgsPostgresResult queryResult;
-    for ( ;; )
-    {
-      queryResult = mConn->PQgetResult();
-      if ( !queryResult.result() )
-        break;
-
-      if ( queryResult.PQresultStatus() != PGRES_TUPLES_OK )
+      lock();
+      if ( mConn->PQsendQuery( fetch ) == 0 ) // fetch features asynchronously
       {
         QgsMessageLog::logMessage( QObject::tr( "Fetching from cursor %1 failed\nDatabase error: %2" ).arg( mCursorName, mConn->PQerrorMessage() ), QObject::tr( "PostGIS" ) );
-        break;
       }
 
-      int rows = queryResult.PQntuples();
-      if ( rows == 0 )
-        continue;
-
-      mLastFetch = rows < mFeatureQueueSize;
-
-      for ( int row = 0; row < rows; row++ )
+      QgsPostgresResult queryResult;
+      for ( ;; )
       {
-        mFeatureQueue.enqueue( QgsFeature() );
-        getFeature( queryResult, row, mFeatureQueue.back() );
-      } // for each row in queue
-    }
-    unlock();
+        queryResult = mConn->PQgetResult();
+        if ( !queryResult.result() )
+          break;
+
+        if ( queryResult.PQresultStatus() != PGRES_TUPLES_OK )
+        {
+          QgsMessageLog::logMessage( QObject::tr( "Fetching from cursor %1 failed\nDatabase error: %2" ).arg( mCursorName, mConn->PQerrorMessage() ), QObject::tr( "PostGIS" ) );
+          break;
+        }
+
+        int rows = queryResult.PQntuples();
+        if ( rows == 0 )
+          continue;
+
+        mLastFetch = rows < mFeatureQueueSize;
+
+        for ( int row = 0; row < rows; row++ )
+        {
+          mFeatureQueue.enqueue( QgsFeature() );
+          getFeature( queryResult, row, mFeatureQueue.back() );
+        } // for each row in queue
+      }
+      unlock();
 
 #if 0 //disabled dynamic queue size
-    if ( timer.elapsed() > 500 && mFeatureQueueSize > 1 )
-    {
-      mFeatureQueueSize /= 2;
-    }
-    else if ( timer.elapsed() < 50 && mFeatureQueueSize < 10000 )
-    {
-      mFeatureQueueSize *= 2;
-    }
+      if ( timer.elapsed() > 500 && mFeatureQueueSize > 1 )
+      {
+        mFeatureQueueSize /= 2;
+      }
+      else if ( timer.elapsed() < 50 && mFeatureQueueSize < 10000 )
+      {
+        mFeatureQueueSize *= 2;
+      }
 #endif
+    }
+
+    if ( mFeatureQueue.empty() )
+    {
+      mLastFetch = true;
+      break;
+    }
+
+    feature = mFeatureQueue.dequeue();
+    mFetched++;
+
+    geometryToDestinationCrs( feature, mTransform );
+    if ( mDistanceWithinEngine && mDistanceWithinEngine->distance( feature.geometry().constGet() ) > mRequest.distanceWithin() )
+    {
+      continue;
+    }
+
+    feature.setValid( true );
+    feature.setFields( mSource->mFields ); // allow name-based attribute lookups
+    return true;
   }
+  QgsDebugMsgLevel( QStringLiteral( "Finished after %1 features" ).arg( mFetched ), 2 );
+  close();
 
-  if ( mFeatureQueue.empty() )
-  {
-    QgsDebugMsg( QStringLiteral( "Finished after %1 features" ).arg( mFetched ) );
-    close();
+  mSource->mShared->ensureFeaturesCountedAtLeast( mFetched );
 
-    mSource->mShared->ensureFeaturesCountedAtLeast( mFetched );
-
-    return false;
-  }
-
-  feature = mFeatureQueue.dequeue();
-  mFetched++;
-
-  feature.setValid( true );
-  feature.setFields( mSource->mFields ); // allow name-based attribute lookups
-  geometryToDestinationCrs( feature, mTransform );
-
-  return true;
+  return false;
 }
 
 bool QgsPostgresFeatureIterator::nextFeatureFilterExpression( QgsFeature &f )
@@ -559,7 +594,7 @@ QString QgsPostgresFeatureIterator::whereClauseRect()
 
   if ( mSource->mRequestedGeomType != QgsWkbTypes::Unknown && mSource->mRequestedGeomType != mSource->mDetectedGeomType )
   {
-    whereClause += QStringLiteral( " AND %1" ).arg( QgsPostgresConn::postgisTypeFilter( mSource->mGeometryColumn, ( QgsWkbTypes::Type )mSource->mRequestedGeomType, castToGeometry ) );
+    whereClause += QStringLiteral( " AND %1" ).arg( QgsPostgresConn::postgisTypeFilter( mSource->mGeometryColumn, mSource->mRequestedGeomType, castToGeometry ) );
   }
 
   QgsDebugMsgLevel( QStringLiteral( "whereClause = %1" ).arg( whereClause ), 4 );
@@ -570,7 +605,10 @@ QString QgsPostgresFeatureIterator::whereClauseRect()
 
 bool QgsPostgresFeatureIterator::declareCursor( const QString &whereClause, long limit, bool closeOnFail, const QString &orderBy )
 {
-  mFetchGeometry = ( !( mRequest.flags() & QgsFeatureRequest::NoGeometry ) || mFilterRequiresGeometry ) && !mSource->mGeometryColumn.isNull();
+  mFetchGeometry = ( !( mRequest.flags() & QgsFeatureRequest::NoGeometry )
+                     || mFilterRequiresGeometry
+                     || ( mRequest.spatialFilterType() == Qgis::SpatialFilterType::DistanceWithin && !mTransform.isShortCircuited() ) )
+                   && !mSource->mGeometryColumn.isNull();
 #if 0
   // TODO: check that all field indexes exist
   if ( !hasAllFields )
@@ -596,8 +634,7 @@ bool QgsPostgresFeatureIterator::declareCursor( const QString &whereClause, long
 
     if ( !mRequest.simplifyMethod().forceLocalOptimization() &&
          mRequest.simplifyMethod().methodType() != QgsSimplifyMethod::NoSimplification &&
-         QgsWkbTypes::flatType( QgsWkbTypes::singleType( usedGeomType ) ) != QgsWkbTypes::Point &&
-         !QgsWkbTypes::isCurvedType( usedGeomType ) )
+         QgsWkbTypes::flatType( QgsWkbTypes::singleType( usedGeomType ) ) != QgsWkbTypes::Point )
     {
       // PostGIS simplification method to use
       QString simplifyPostgisMethod;
@@ -606,7 +643,7 @@ bool QgsPostgresFeatureIterator::declareCursor( const QString &whereClause, long
       bool postSimplification;
       postSimplification = false; // default to false. Set to true only for PostGIS >= 2.2 when using st_removerepeatedpoints
 
-      if ( mRequest.simplifyMethod().methodType() == QgsSimplifyMethod::OptimizeForRendering )
+      if ( !QgsWkbTypes::isCurvedType( usedGeomType ) && mRequest.simplifyMethod().methodType() == QgsSimplifyMethod::OptimizeForRendering )
       {
         // Optimize simplification for rendering
         if ( mConn->majorVersion() < 2 )
@@ -625,7 +662,7 @@ bool QgsPostgresFeatureIterator::declareCursor( const QString &whereClause, long
             // For postgis >= 2.2 Use ST_RemoveRepeatedPoints instead
             // Do it only if threshold is <= 1 pixel to avoid holes in adjacent polygons
             // We should perhaps use it always for Linestrings, even if threshold > 1 ?
-            if ( mRequest.simplifyMethod().threshold() <= 1.0 )
+            if ( mRequest.simplifyMethod().threshold() <= 1.0f )
             {
               simplifyPostgisMethod = QStringLiteral( "st_removerepeatedpoints" );
               postSimplification = true; // Ask to apply a post-filtering simplification
@@ -645,10 +682,10 @@ bool QgsPostgresFeatureIterator::declareCursor( const QString &whereClause, long
           simplifyPostgisMethod = QStringLiteral( "st_simplifypreservetopology" );
         }
       }
-      QgsDebugMsg(
-        QString( "PostGIS Server side simplification : threshold %1 pixels - method %2" )
+      QgsDebugMsgLevel(
+        QStringLiteral( "PostGIS Server side simplification : threshold %1 pixels - method %2" )
         .arg( mRequest.simplifyMethod().threshold() )
-        .arg( simplifyPostgisMethod )
+        .arg( simplifyPostgisMethod ), 3
       );
 
       geom = QStringLiteral( "%1(%2,%3)" )
@@ -693,7 +730,7 @@ bool QgsPostgresFeatureIterator::declareCursor( const QString &whereClause, long
       break;
 
     case PktFidMap:
-      Q_FOREACH ( int idx, mSource->mPrimaryKeyAttrs )
+      for ( int idx : std::as_const( mSource->mPrimaryKeyAttrs ) )
       {
         query += delim + mConn->fieldExpression( mSource->mFields.at( idx ) );
         delim = ',';
@@ -758,7 +795,7 @@ bool QgsPostgresFeatureIterator::getFeature( QgsPostgresResult &queryResult, int
       memcpy( &wkbType, featureGeom + 1, sizeof( wkbType ) );
       QgsWkbTypes::Type newType = QgsPostgresConn::wkbTypeFromOgcWkbType( wkbType );
 
-      if ( ( unsigned int )newType != wkbType )
+      if ( static_cast< unsigned int >( newType ) != wkbType )
       {
         // overwrite type
         unsigned int n = newType;
@@ -854,8 +891,9 @@ bool QgsPostgresFeatureIterator::getFeature( QgsPostgresResult &queryResult, int
     case PktFidMap:
     {
       QVariantList primaryKeyVals;
+      primaryKeyVals.reserve( mSource->mPrimaryKeyAttrs.size() );
 
-      Q_FOREACH ( int idx, mSource->mPrimaryKeyAttrs )
+      for ( int idx : std::as_const( mSource->mPrimaryKeyAttrs ) )
       {
         QgsField fld = mSource->mFields.at( idx );
         QVariant v;
@@ -882,7 +920,7 @@ bool QgsPostgresFeatureIterator::getFeature( QgsPostgresResult &queryResult, int
     break;
 
     case PktUnknown:
-      Q_ASSERT( !"FAILURE: cannot get feature with unknown primary key" );
+      Q_ASSERT_X( false, "QgsPostgresFeatureIterator::getFeature", "FAILURE: cannot get feature with unknown primary key" );
       return false;
   }
 
