@@ -219,13 +219,17 @@ struct VTable
       }
 
       QgsVectorDataProvider *provider = mLayer ? mLayer->dataProvider() : mProvider;
-      if ( provider->wkbType() != QgsWkbTypes::NoGeometry )
+
+      // spatialite doesn't support curved geometries, it will be converted to linear in qgsGeometryToSpatialiteBlob
+      QgsWkbTypes::Type layerType = QgsWkbTypes::linearType( provider->wkbType() );
+
+      if ( layerType != QgsWkbTypes::NoGeometry )
       {
         // we have here a convenient hack
         // the type of a column can be declared with two numeric arguments, usually for setting numeric precision
         // we are using them to set the geometry type and srid
         // these will be reused by the provider when it will introspect the query to detect types
-        sqlFields << QStringLiteral( "geometry geometry(%1,%2)" ).arg( provider->wkbType() ).arg( provider->crs().postgisSrid() );
+        sqlFields << QStringLiteral( "geometry geometry(%1,%2)" ).arg( layerType ).arg( provider->crs().postgisSrid() );
 
         // add a hidden field for rtree filtering
         sqlFields << QStringLiteral( "_search_frame_ HIDDEN BLOB" );
@@ -237,7 +241,7 @@ struct VTable
         mPkColumn = pkAttributeIndexes.at( 0 );
       }
 
-      mCreationStr = "CREATE TABLE vtable (" + sqlFields.join( QStringLiteral( "," ) ) + ")";
+      mCreationStr = "CREATE TABLE vtable (" + sqlFields.join( QLatin1Char( ',' ) ) + ")";
 
       mCrs = provider->crs().postgisSrid();
     }
@@ -691,11 +695,13 @@ int vtableColumn( sqlite3_vtab_cursor *cursor, sqlite3_context *ctxt, int idx )
     switch ( v.type() )
     {
       case QVariant::Int:
-      case QVariant::UInt:
       case QVariant::Bool:
+        // read signed integer
         sqlite3_result_int( ctxt, v.toInt() );
         break;
+      case QVariant::UInt:
       case QVariant::LongLong:
+        // read 64 bits signed integer (or 32 bits unsigned one)
         sqlite3_result_int64( ctxt, v.toLongLong() );
         break;
       case QVariant::Double:
@@ -857,7 +863,8 @@ void registerQgisFunctions( sqlite3 *db )
   QStringList reservedFunctions;
   reservedFunctions << QStringLiteral( "left" ) << QStringLiteral( "right" ) << QStringLiteral( "union" );
   // register QGIS expression functions
-  Q_FOREACH ( QgsExpressionFunction *foo, QgsExpression::Functions() )
+  const QList<QgsExpressionFunction *> functions = QgsExpression::Functions();
+  for ( QgsExpressionFunction *foo : functions )
   {
     if ( foo->usesGeometry( nullptr ) || foo->lazyEval() )
     {
@@ -879,11 +886,11 @@ void registerQgisFunctions( sqlite3 *db )
       params = -1;
     }
 
-    Q_FOREACH ( QString name, names ) // for each alias
+    for ( QString name : std::as_const( names ) ) // for each alias
     {
       if ( reservedFunctions.contains( name ) ) // reserved keyword
         name = "_" + name;
-      if ( name.startsWith( QLatin1String( "$" ) ) )
+      if ( name.startsWith( QLatin1Char( '$' ) ) )
         continue;
 
       // register the function and pass the pointer to the Function* as user data
