@@ -150,6 +150,11 @@ int QgsWmsRenderContext::tileBuffer() const
   return tileBuffer;
 }
 
+bool QgsWmsRenderContext::renderMapTiles() const
+{
+  return QgsServerProjectUtils::wmsRenderMapTiles( *mProject );
+}
+
 int QgsWmsRenderContext::precision() const
 {
   int precision = QgsServerProjectUtils::wmsFeatureInfoPrecision( *mProject );
@@ -207,7 +212,7 @@ QStringList QgsWmsRenderContext::flattenedQueryLayers( const QStringList &layerN
     return _result;
   };
 
-  for ( const auto &name : qgis::as_const( layerNames ) )
+  for ( const auto &name : std::as_const( layerNames ) )
   {
     result.append( findLeaves( name ) );
   }
@@ -379,7 +384,7 @@ void QgsWmsRenderContext::initRestrictedLayers()
   QStringList restrictedLayersNames;
   QgsLayerTreeGroup *root = mProject->layerTreeRoot();
 
-  for ( const QString &l : qgis::as_const( restricted ) )
+  for ( const QString &l : std::as_const( restricted ) )
   {
     const QgsLayerTreeGroup *group = root->findGroup( l );
     if ( group )
@@ -513,7 +518,7 @@ void QgsWmsRenderContext::searchLayersToRenderStyle()
 
     if ( ! param.mExternalUri.isEmpty() && ( mFlags & AddExternalLayers ) )
     {
-      std::unique_ptr<QgsMapLayer> layer = qgis::make_unique< QgsRasterLayer >( param.mExternalUri, param.mNickname, QStringLiteral( "wms" ) );
+      std::unique_ptr<QgsMapLayer> layer = std::make_unique< QgsRasterLayer >( param.mExternalUri, param.mNickname, QStringLiteral( "wms" ) );
 
       if ( layer->isValid() )
       {
